#include "dumpwtb.h"
#include "wtruntime/version.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void print_help(const char* progname)
{
	printf("Usage: %s OPTION... FILE\n", progname);
	printf("\n");
	printf("    -a    Show all information\n");
	printf("    -h    Show header\n");
	printf("    -s    Show section headers\n");
	printf("    -t    Show thread headers\n");
	printf("    -m    Show metadata\n");
	printf("    -b    Show byte-order-swap informations\n");
	printf("    -c    Disassemble code\n");
	printf("    -d    Show data section\n");
	printf("    -I    Interpret data section based on its BOS section\n");
	printf("    -z    Assume zero terminated strings in data sections\n");
	printf("    -L    List debug labels\n");
	printf("    -l    Show labels in disassembly\n");
	printf("    -i    Show imports\n");
	printf("    -e    Show exports\n");
	printf("    -g    Display integrity sections\n");
}

void print_version(const char* progname)
{
	printf("ReadWTB is part of the Welltype suite v%u.%u.%u\n",
		WT_GETMAJOR(WTR_VERSION),
		WT_GETMINOR(WTR_VERSION),
		WT_GETPATCH(WTR_VERSION));
}

int main(int argc, char* argv[])
{
	int            argi      = 1;
	wt_u32         mode      = 0;
	const char*    filename  = NULL;
	FILE*          file      = NULL;
	void*          data      = NULL;
	void*          allocated = NULL;
	unsigned char* dptr      = NULL;
	wt_u32         size      = 0;
	int            res       = 10;

	for(;argi<argc;++argi)
	{
		if('-'==argv[argi][0])
		{
			const char* opt = argv[argi] + 1;

			if(0==strcmp("--help", argv[argi])) { print_help(argv[0]); return 0; }
			if(0==strcmp("--version", argv[argi])) { print_version(argv[0]); return 0; }

			for(;*opt;++opt)
			{
				switch(*opt)
				{
				case 'a': mode  = DWM__ALL;        break;
				case 'h': mode |= DWM_HEADER;      break;
				case 's': mode |= DWM_SECTIONS;    break;
				case 't': mode |= DWM_THREADS;     break;
				case 'm': mode |= DWM_METADATA;    break;
				case 'b': mode |= DWM_BOS;         break;
				case 'c': mode |= DWM_CODE;        break;
				case 'd': mode |= DWM_DATA;        break;
				case 'I': mode |= DWM_DATABOSINTR; break;
				case 'z': mode |= DWM_ASSUMESTR;   break;
				case 'L': mode |= DWM_SHOWDLABELS; break;
				case 'l': mode |= DWM_CODELABELS;  break;
				case 'i': mode |= DWM_IMPORT;      break;
				case 'e': mode |= DWM_EXPORT;      break;
				case 'g': mode |= DWM_INTEGRITY;   break;

				default:
					{
						fprintf(stderr, "Unknown argument: \"%s\"\n",
							argv[argi]);
						return 1;
					}
					break;
				}
			}
		}
		else
		if(NULL==filename)
		{
			filename = argv[argi];
		}
		else
		{
			fprintf(stderr, "Too many arguments.\n");
			return 1;
		}
	}

	if(NULL==filename || 0==mode)
	{
		print_help(argv[0]);
		return 1;
	}

	if(NULL==(file=fopen(filename, "rb")))
	{
		fprintf(stderr, "%s: %s: No such file or directory.\n", argv[0],
			filename);
		return 1;
	}

	fseek(file, 0, SEEK_END);
	size = (wt_u32)ftell(file);
	fseek(file, 0, SEEK_SET);

	if(NULL==(data=allocated=malloc(size+16)))
	{
		fclose(file);
		fprintf(stderr, "%s: %s: Out of memory.\n", argv[0], filename);
		return 1;
	}

	data = (void*)((((wt_uintptr)data) + 15) & (-16));
	if(1!=fread(data, size, 1, file))
	{
		fclose(file);
		fprintf(stderr, "%s: %s: I/O error.\n", argv[0], filename);
		return 1;
	}

	fclose(file);

	dptr = (unsigned char*)data;
	if(size>2 && 0==memcmp(dptr, "#!", 2))
	{
		size -= 2;
		dptr += 2;

		/* skip the shebang (find the magic) */

		while(size>sizeof(wt_u32))
		{
			if(WTBMAG0==dptr[0] && WTBMAG1==dptr[1] && WTBMAG2==dptr[2] && WTBMAG3==dptr[3])
			{
				break;
			}

			++dptr;
			--size;
		}
	}

	res = dumpwtb(dptr, size, mode);

	free(allocated);

	return res;
}
