#include "config_private.h"
#include "type.h"
#include "program.h"
#include "profile.h"
#include "byteorder.h"
#include <string.h>
#if WTRUNTIME_DEBUG
#  include <assert.h>
#endif /* WTRUNTIME_DEBUG */

#define ALIGNMENT   8
#define ROUNDUP(n)  ((((wt_uintptr)(n)) + ALIGNMENT - 1) & (-ALIGNMENT))
#define ROUNDUP4(n) ((((wt_uintptr)(n)) + 3) & (-4))

void wtr_type_release(struct wtr_type* type)
{
	struct wtr_profile* const profile = type->program->profile;

	memset(type, 0, sizeof(struct wtr_type));
	wtr_profile_vm_free(profile, type);
}

static wt_bool wtr_type_is_equal_impl(const struct wtr_type* lhs, const struct wtr_type* rhs, wt_bool with_preserved, wt_u32 preserved_rhs_local_typeid)
{
	if(lhs->kind!=rhs->kind)
	{
		return WT_FALSE;
	}

	if(0!=strcmp(lhs->name, rhs->name))
	{
		return WT_FALSE;
	}

	switch(lhs->kind)
	{
	case WTBSYMT_RECORD:
		{
			const struct wtr_type_record* lrec = (const struct wtr_type_record*)lhs;
			const struct wtr_type_record* rrec = (const struct wtr_type_record*)rhs;
			wt_u32                        i;

			if(lrec->fldc!=rrec->fldc)
			{
				return WT_FALSE;
			}

			for(i=0;i<lrec->fldc;++i)
			{
				if(0!=strcmp(lrec->fldv[i].name, rrec->fldv[i].name))
				{
					return WT_FALSE;
				}

				if(lrec->fldv[i].type!=rrec->fldv[i].type)
				{
					if(!with_preserved) { return WT_FALSE; }

					if(!(rrec->fldv[i].type & WTTF_LOCAL) || preserved_rhs_local_typeid!=(rrec->fldv[i].type & WTTM_USERTYPEID))
					{
						return WT_FALSE;
					}
				}
			}
		}
		break;

	case WTBSYMT_DATATYPE:
		{
			const struct wtr_type_datatype* ldt = (const struct wtr_type_datatype*)lhs;
			const struct wtr_type_datatype* rdt = (const struct wtr_type_datatype*)rhs;
			wt_u32                          i;

			if(ldt->numcons!=rdt->numcons)
			{
				return WT_FALSE;
			}

			if(ldt->defcons!=rdt->defcons)
			{
				return WT_FALSE;
			}

			for(i=0;i<ldt->numcons;++i)
			{
				const struct wtr_type_datacons** lcons = (const struct wtr_type_datacons**)ldt->conses;
				const struct wtr_type_datacons** rcons = (const struct wtr_type_datacons**)rdt->conses;
				wt_u32                    typc;
				wt_u32                    j;

				if(0!=strcmp((*lcons)->name, (*rcons)->name))
				{
					return WT_FALSE;
				}

				if((*lcons)->typc!=(*rcons)->typc)
				{
					return WT_FALSE;
				}

				typc = (*lcons)->typc;

				for(j=0;j<typc;++j)
				{
					if((*lcons)->typv[j]!=(*rcons)->typv[j])
					{
						if(!with_preserved) { return WT_FALSE; }

						if(!((*rcons)->typv[j] & WTTF_LOCAL) || preserved_rhs_local_typeid!=((*rcons)->typv[j] & WTTM_USERTYPEID))
						{
							return WT_FALSE;
						}
					}
				}
			}
		}
		break;

	case WTBSYMT_FUNCTIONTYPE:
		{
			const struct wtr_type_functype* lft = (const struct wtr_type_functype*)lhs;
			const struct wtr_type_functype* rft = (const struct wtr_type_functype*)rhs;
			wt_u32                          i;

			if(lft->argc!=rft->argc)
			{
				return WT_FALSE;
			}

			if(lft->retc!=rft->retc)
			{
				return WT_FALSE;
			}

			if(lft->is_pure!=rft->is_pure)
			{
				return WT_FALSE;
			}

			for(i=0;i<lft->argc;++i)
			{
				if(lft->argv[i]!=rft->argv[i])
				{
					if(!with_preserved) { return WT_FALSE; }

					if(!(rft->argv[i] & WTTF_LOCAL) || preserved_rhs_local_typeid!=(rft->argv[i] & WTTM_USERTYPEID))
					{
						return WT_FALSE;
					}
				}
			}

			for(i=0;i<lft->retc;++i)
			{
				if(lft->retv[i]!=rft->retv[i])
				{
					if(!with_preserved) { return WT_FALSE; }

					if(!(rft->retv[i] & WTTF_LOCAL) || preserved_rhs_local_typeid!=(rft->retv[i] & WTTM_USERTYPEID))
					{
						return WT_FALSE;
					}
				}
			}
		}
		break;

	case WTBSYMT_ENUM:
		{
			const struct wtr_type_enum* le = (const struct wtr_type_enum*)lhs;
			const struct wtr_type_enum* re = (const struct wtr_type_enum*)rhs;
			wt_u32               i;

			if(le->enumeratorc!=re->enumeratorc)
			{
				return WT_FALSE;
			}

			for(i=0;i<le->enumeratorc;++i)
			{
				if(0!=strcmp(le->enumeratorv[i], re->enumeratorv[i]))
				{
					return WT_FALSE;
				}
			}
		}
		break;

	case WTBSYMT_PRIVATERECORD:
	case WTBSYMT_LIMITEDRECORD:
		break;

	default:
#if WTRUNTIME_DEBUG
		assert(! "unimplemented type");
#endif /* WTRUNTIME_DEBUG */
		return WT_FALSE;
	}

	return WT_TRUE;
}

wt_bool wtr_type_is_equal(const struct wtr_type* lhs, const struct wtr_type* rhs)
{
	return wtr_type_is_equal_impl(lhs, rhs, WT_FALSE, 0);
}

wt_bool wtr_type_is_equal_preserved(const struct wtr_type* lhs, const struct wtr_type* rhs, wt_u32 preserved_rhs_local_typeid)
{
	return wtr_type_is_equal_impl(lhs, rhs, WT_TRUE, preserved_rhs_local_typeid);
}

void wtr_type_post_resolve(struct wtr_type* type, wt_u32 preserved_typeid, wt_u32 resolved_typeid)
{
	switch(type->kind)
	{
	case WTBSYMT_DATATYPE:
		{
			struct wtr_type_datatype*  dt   = (struct wtr_type_datatype*)type;
			struct wtr_type_datacons** cons = dt->conses;
			wt_u32                     i;

			for(i=0;i<dt->numcons;++cons,++i)
			{
				wt_u32* ptr = (*cons)->typv;
				wt_u32* end = ptr + (*cons)->typc;

				for(;ptr<end;++ptr)
				{
					if(((*ptr)&WTTF_LOCAL) && preserved_typeid==((*ptr)&WTTM_USERTYPEID))
					{
						*ptr = resolved_typeid | (preserved_typeid & (WTTF_REF|WTTF_MUTABLE|WTTF_SEQ));
					}
				}
			}
		}
		break;
	}
}

struct wtr_type* wtr_type_record_new(struct wtr_program* program, const char* name, wt_u32 fldc, const wt_u32* fldv, wt_bo bo)
{
	struct wtr_type_record* type  = NULL;
	const char*             names = (const char*)(fldv + fldc);
	wt_u32                  nlen  = strlen(name) + 1;
	wt_u32                  nsize = 0;

	{
		const char* nameptr = names;
		wt_u32      i;

		for(i=0;i<fldc;++i)
		{
			while(*nameptr)
			{
				++nameptr;
			}

			++nameptr;
		}

		nsize = nameptr - names;
	}

	type = (struct wtr_type_record*)wtr_profile_vm_malloc(program->profile, sizeof(struct wtr_type_record)+fldc*sizeof(struct wtr_type_field)+nlen+nsize);

	if(type)
	{
		struct wtr_type_field* fptr  = (struct wtr_type_field*)(type + 1);
		char*                  after = (char*)(fptr + fldc);

		strcpy(after, name);

		type->type.name     = after;
		type->type.id       = 0;
		type->type.kind     = WTBSYMT_RECORD;
		type->type.refcount = 0;
		type->type.program  = program;
		type->fldc          = fldc;
		type->fldv          = fptr;

		after += nlen;

		for(;fldc;++fptr,++fldv,--fldc)
		{
			fptr->name = after;
			fptr->type = wtr_botoh32(*fldv, bo);

			while(*names)
			{
				*after++ = *names++;
			}

			*after++ = *names++;

			if(!wtr_program_resolvetype(program, &fptr->type))
			{
				wtr_profile_vm_free(program->profile, type);
				return NULL;
			}
		}
	}

	return (struct wtr_type*)type;
}

struct wtr_type* wtr_type_privaterecord_new(struct wtr_program* program, const char* name)
{
	const wt_u32                   nlen = strlen(name) + 1;
	struct wtr_type_privaterecord* type = (struct wtr_type_privaterecord*)wtr_profile_vm_malloc(program->profile, sizeof(struct wtr_type_privaterecord)+nlen);

	if(type)
	{
		char* after = (char*)(type + 1);

		type->type.name     = after;
		type->type.id       = 0;
		type->type.kind     = WTBSYMT_PRIVATERECORD;
		type->type.refcount = 0;
		type->type.program  = program;

		strcpy(after, name);
	}

	return (struct wtr_type*)type;
}

struct wtr_type* wtr_type_limitedrecord_new(struct wtr_program* program, const char* name)
{
	const wt_u32                   nlen = strlen(name) + 1;
	struct wtr_type_limitedrecord* type = (struct wtr_type_limitedrecord*)wtr_profile_vm_malloc(program->profile, sizeof(struct wtr_type_limitedrecord)+nlen);

	if(type)
	{
		char* after = (char*)(type + 1);

		type->type.name     = after;
		type->type.id       = 0;
		type->type.kind     = WTBSYMT_LIMITEDRECORD;
		type->type.refcount = 0;
		type->type.program  = program;

		strcpy(after, name);
	}

	return (struct wtr_type*)type;
}

struct wtr_type* wtr_type_datatype_new(struct wtr_program* program, const char* name, const wt_u32 preserve_typeid, const void* data, wt_bo bo)
{
	struct wtr_type_datatype* type = NULL;
	const wt_u32*             raw  = (const wt_u32*)data;
	const wt_u32*             sub  = NULL;

	/* populate required size */

	wt_u32 size    = sizeof(struct wtr_type_datatype);
	wt_u32 nlen    = ROUNDUP(strlen(name) + 1);
	wt_u32 numcons = wtr_botoh32(*raw++, bo);
	wt_u32 defcons = wtr_botoh32(*raw++, bo);
	wt_u32 i;

	if(defcons>numcons)
	{
		/* the id of the default constructor cannot be greater than the number of constructors */
		return NULL;
	}

	size += nlen;
	size += sizeof(struct wtr_type_datacons*)*numcons;
	size += ROUNDUP(sizeof(struct wtr_type_datacons))*numcons;

	for(sub=raw,i=0;i<numcons;++i)
	{
		wt_u32 typc;
		wt_u32 clen = ROUNDUP4(strlen((const char*)sub) + 1);

		sub = (const wt_u32*)(((wt_uintptr)sub) + clen);

		typc = wtr_botoh32(*sub++, bo);
		sub += typc;

		size += clen + sizeof(wt_u32)*(typc + 1);
	}

	/* allocate memory */

	type = (struct wtr_type_datatype*)wtr_profile_vm_malloc(program->profile, size);

	if(type)
	{
		char*   after = (char*)(type + 1) + sizeof(struct wtr_type_datacons*)*numcons;
		wt_u32* ptr   = (wt_u32*)(after + nlen);

		type->type.name     = after;
		type->type.id       = 0;
		type->type.kind     = WTBSYMT_DATATYPE;
		type->type.refcount = 0;
		type->type.program  = program;
		type->numcons       = numcons;
		type->defcons       = defcons;

		strcpy(after, name);

		for(sub=raw,i=0;i<numcons;++i)
		{
			wt_u32                    typc;
			wt_u32                    clen  = ROUNDUP4(strlen((const char*)sub) + 1);
			char*                     cname = (char*)ptr;
			struct wtr_type_datacons* cons;

			ptr  = (wt_u32*)ROUNDUP(((wt_uintptr)ptr) + clen);
			cons = (struct wtr_type_datacons*)ptr;
			ptr  = cons->typv;

			strcpy(cname, (const char*)sub);

			sub  = (const wt_u32*)(((wt_uintptr)sub) + clen);
			typc = wtr_botoh32(*sub++, bo);

			type->conses[i] = cons;
			cons->name      = cname;
			cons->typc      = typc;

			while(typc--)
			{
				*ptr = wtr_botoh32(*sub++, bo);

				if((!((*ptr)&WTTF_LOCAL) || preserve_typeid!=((*ptr)&WTTM_USERTYPEID)) && !wtr_program_resolvetype(program, ptr))
				{
					wtr_profile_vm_free(program->profile, type);
					return NULL;
				}

				++ptr;
			}
		}
	}

	return (struct wtr_type*)type;
}

struct wtr_type* wtr_type_functype_new(struct wtr_program* program, const char* name, const wt_u32 preserve_typeid, const void* data, wt_bo bo)
{
	struct wtr_type_functype* type = NULL;
	wt_u32                    nlen = ROUNDUP(strlen(name) + 1);
	wt_u32                    size = sizeof(struct wtr_type_functype) + nlen;
	const wt_u32*             raw  = (const wt_u32*)data;
	wt_u32                    argc = wtr_botoh32(*raw++, bo);
	wt_u32                    retc = wtr_botoh32(raw[argc], bo);

	size += sizeof(wt_u32) * (argc + retc);

	type = (struct wtr_type_functype*)wtr_profile_vm_malloc(program->profile, size);

	if(type)
	{
		char*   after = (char*)(type + 1);
		wt_u32* arr   = (wt_u32*)(after + nlen);
		wt_u32  i;

		type->type.name     = after;
		type->type.id       = 0;
		type->type.kind     = WTBSYMT_FUNCTIONTYPE;
		type->type.refcount = 0;
		type->type.program  = program;
		type->argc          = argc;
		type->retc          = retc;
		type->argv          = arr;
		type->retv          = arr + argc;

		strcpy(after, name);

		for(i=0;i<argc;++i)
		{
			type->argv[i] = wtr_botoh32(*raw++, bo);
		}

		++raw;

		for(i=0;i<retc;++i)
		{
			type->retv[i] = wtr_botoh32(*raw++, bo);
		}

		type->is_pure = *(const wt_u8*)raw;
	}

	return (struct wtr_type*)type;
}

struct wtr_type* wtr_type_enum_new(struct wtr_program* program, const char* name, const void* data, wt_bo bo)
{
	struct wtr_type_enum* type = NULL;
	wt_u32                nlen = strlen(name) + 1;
	wt_u32                size = ROUNDUP(sizeof(struct wtr_type_enum)) + nlen;
	const char*           raw  = (const char*)data;
	wt_u32                cnt0 = wtr_botoh32(*(const wt_u32*)raw, bo);

	raw  += sizeof(wt_u32);
	size += sizeof(char*) * cnt0;

	{
		const char* base = raw;
		wt_u32      cnt  = cnt0;

		while(cnt--)
		{
			wt_u32 elen = strlen(base) + 1;
			size += elen;
			base += elen;
		}
	}

	type = (struct wtr_type_enum*)wtr_profile_vm_malloc(program->profile, size);

	if(type)
	{
		const char** ev    = (const char**)(ROUNDUP((wt_uintptr)(type + 1)));
		char*        after = (char*)(ev + cnt0);
		char*        ptr   = after + nlen;
		wt_u32       idx   = 0;

		type->type.name     = after;
		type->type.id       = 0;
		type->type.kind     = WTBSYMT_ENUM;
		type->type.refcount = 0;
		type->type.program  = program;
		type->enumeratorc   = cnt0;
		type->enumeratorv   = ev;

		strcpy(after, name);

		while(cnt0--)
		{
			wt_u32 elen = strlen(raw) + 1;
			type->enumeratorv[idx++] = ptr;
			strcpy(ptr, raw);
			ptr += elen;
			raw += elen;
		}
	}

	return (struct wtr_type*)type;
}
